from django.db import models
from django.urls import reverse


class FancyModel(models.Model):
    class Meta:
        app_label = 'testapp'

    def get_absolute_url():
        return reverse('model-view-redirect')



from dataclasses import dataclass
from django.contrib.auth.models import AnonymousUser, User
from django.test import RequestFactory, SimpleTestCase, override_settings

from cravensworth.core.models import (
    Allocation,
    Audience,
    Context,
    Experiment,
)
from cravensworth.core.providers import DjangoRequestContextProvider
from cravensworth.core.experiment import (
    DEFAULT_CRAVENSWORTH_COOKIE,
    _CravensworthState,
    _extract_overrides,
)


class TestContext(SimpleTestCase):
    def test_random_identity(self):
        context = Context()
        identity = context.identity('random', None)
        self.assertTrue(0 <= identity < 100)

    def test_random_identity_cached(self):
        context = Context()
        identity1 = context.identity('random', None)
        identity2 = context.identity('random', None)
        self.assertEqual(identity1, identity2)

    @dataclass
    class ValueObject:
        value: str

    def test_valid_idenity_keypath_value(self):
        context = Context(
            {
                'a': '123',
                'b': {'c': 456, 'd': self.ValueObject(value='a')},
                'c': {3},
                'd': 42,
                'e': self.ValueObject(value='z'),
            }
        )
        keys = ['a', 'b.c', 'b.d.value', 'c', 'd', 'e.value']
        for key in keys:
            context.identity(key, None)  # Does not raise KeyError

    def test_idenity_raises_if_value_not_found(self):
        keys = ['j', 'a.z', 'b.y', 'b.c.x']
        context = Context(
            {
                'a': '123',
                'b': {'c': '456'},
            }
        )
        for key in keys:
            with self.assertRaises(KeyError):
                context.identity(key, None)

    def test_identities_differ_when_seeded(self):
        context = Context({'a': '123'})
        identity1 = context.identity('a', 'seed1')
        identity2 = context.identity('a', 'seed2')
        self.assertNotEqual(identity1, identity2)


class TestDjangoContextProvider(SimpleTestCase):
    factory = RequestFactory()
    provider = DjangoRequestContextProvider()

    def test_request_missing(self):
        with self.assertRaisesRegex(ValueError, 'request is required'):
            self.provider.context()

    def test_populates_anonymous_request(self):
        request = self.factory.get('/')
        request.user = AnonymousUser()
        context = self.provider.context(request=request)

        user = context.get('user')
        tracking_key = context.get('anonymous')
        self.assertTrue(user is not None and user.is_anonymous)
        self.assertTrue(tracking_key is not None and tracking_key != '')

    def test_populates_authenticated_request(self):
        request = self.factory.get('/')
        request.user = User()
        context = self.provider.context(request=request)

        user = context.get('user')
        tracking_key = context.get('anonymous')
        self.assertTrue(user is not None and not user.is_anonymous)
        self.assertTrue(tracking_key is not None and tracking_key != '')


class TestAllocation(SimpleTestCase):
    def test_validate_valid_allocation(self):
        Allocation(variant='blue', percent=0).validate()
        Allocation(variant='red', percent=20).validate()
        Allocation(variant='green', percent=100).validate()

    def test_validate_invalid_variant_name_raises(self):
        for variant in [*r' ~!@#$%^&*()`-+=[]{};:<>,./?"', "'", '']:
            with self.assertRaisesRegex(
                ValueError, 'Variant must contain only'
            ):
                Allocation(variant=variant, percent=50).validate()

    def test_validate_negative_percent_raises(self):
        with self.assertRaisesRegex(ValueError, 'negative'):
            Allocation(variant='blue', percent=-20).validate()

    def test_validate_percent_too_big_raises(self):
        with self.assertRaisesRegex(ValueError, 'greater than 100'):
            Allocation(variant='blue', percent=110).validate()


class TestAudience(SimpleTestCase):
    def test_validate_rule_syntax_error(self):
        with self.assertRaisesRegex(ValueError, 'Invalid rule'):
            Audience(
                rule=')()()(',
                allocations=[
                    Allocation('blue', percent=100),
                ],
            ).validate()

    def test_validate_valid_rule_syntax(self):
        Audience(
            rule='user.id == 12',
            allocations=[
                Allocation('blue', percent=100),
            ],
        ).validate()

    def test_validate_allocations_less_than_100_percent(self):
        with self.assertRaisesRegex(ValueError, '100 percent'):
            Audience(
                rule=None,
                allocations=[
                    Allocation('blue', percent=10),
                    Allocation('red', percent=10),
                ],
            ).validate()

    def test_validate_allocations_greater_than_100_percent(self):
        with self.assertRaisesRegex(ValueError, '100 percent'):
            Audience(
                rule=None,
                allocations=[
                    Allocation('red', percent=11),
                    Allocation('green', percent=90),
                ],
            ).validate()

    def test_matches_default_rule(self):
        audience = Audience(
            rule=None,
            allocations=[
                Allocation('red', percent=100),
            ],
        ).validate()
        self.assertTrue(audience.matches({}))

    def test_matches_raises_for_non_bool_result(self):
        audience = Audience(
            rule='1 + 1',
            allocations=[
                Allocation('red', percent=100),
            ],
        )
        with self.assertRaises(TypeError):
            self.assertTrue(audience.matches({}))

    def test_matches_trivial_rules(self):
        rules = [
            ('True', True),
            ('not False', True),
            ('1 == 1', True),
            ('2 != 5', True),
            ('10 % 3 == 1', True),
            ('(1 + 2) * 5 == 15', True),
            ('200 > 199', True),
            ('1 >= 0', True),
            ('0 >= 0', True),
            ('42 < 45', True),
            ('44 <= 555', True),
            ('False', False),
            ('not True', False),
            ('1 != 1', False),
            ('2 == 5', False),
            ('10 % 3 == 2', False),
            ('(1 + 2) * 5 == 11', False),
            ('200 < 199', False),
            ('1 <= 0', False),
            ('42 > 45', False),
            ('44 >= 555', False),
        ]
        for rule, expected in rules:
            audience = Audience(
                rule,
                allocations=[
                    Allocation('red', percent=100),
                ],
            ).validate()
            self.assertEqual(audience.matches({}), expected)

    def test_matches_fancier_rules(self):
        rules = [
            ('name == "bojack"', {'name': 'bojack'}, True),
            ('name == "bojack"', {'name': 'sebastian'}, False),
            (
                'lang == "en" or country == "US"',
                {'lang': 'en', 'country': 'CA'},
                True,
            ),
            (
                'lang == "en" or country == "US"',
                {'lang': 'fr', 'country': 'CA'},
                False,
            ),
            (
                'lang == "en" and country == "US"',
                {'lang': 'en', 'country': 'US'},
                True,
            ),
            ('is_mobile', {'is_mobile': True}, True),
            ('is_mobile', {'is_mobile': False}, False),
            (
                'firstname != lastname',
                {'firstname': 'chad', 'lastname': 'chadington'},
                True,
            ),
            ('user.id > 100', {'user': {'id': 100}}, False),
            ('user.id > 100', {'user': {'id': 101}}, True),
            ('user["id"] > 100', {'user': {'id': 100}}, False),
            ('user["id"] > 100', {'user': {'id': 101}}, True),
            (
                'user.is_authenticated',
                {'user': {'is_authenticated': True}},
                True,
            ),
            ('not is_admin', {'is_admin': False}, True),
            ('not is_admin', {'is_admin': True}, False),
            (
                '(lang == "en" and country == "US") or is_admin',
                {'lang': 'en', 'country': 'FR', 'is_admin': True},
                True,
            ),
            (
                '(lang == "en" and country == "US") or is_admin',
                {'lang': 'en', 'country': 'US', 'is_admin': False},
                True,
            ),
            (
                '(lang == "en" and country == "US") or is_admin',
                {'lang': 'en', 'country': 'FR', 'is_admin': False},
                False,
            ),
            ('score >= 90', {'score': 95}, True),
            ('score >= 90', {'score': 85}, False),
            ('18 <= age < 65', {'age': 30}, True),
            ('18 <= age < 65', {'age': 70}, False),
            ('lang in ["en", "fr"]', {'lang': 'en'}, True),
            ('lang in ["en", "fr"]', {'lang': 'de'}, False),
            ('lang in {"en", "fr"}', {'lang': 'en'}, True),
            ('"lang" in {"lang": "de"}', {}, True),
            ('"lang" in properties', {'properties': {'lang': 'de'}}, True),
            ('"country" in properties', {'properties': {'lang': 'de'}}, False),
            ('balance - debt > 0', {'balance': 100, 'debt': 20}, True),
            ('balance - debt > 0', {'balance': 10, 'debt': 20}, False),
            (
                'profile.active and profile.email_verified',
                {'profile': {'active': True, 'email_verified': False}},
                False,
            ),
            (
                'profile.active and profile.email_verified',
                {'profile': {'active': True, 'email_verified': True}},
                True,
            ),
            ('tags.count > 0', {'tags': {'count': 2}}, True),
            ('tags.count > 0', {'tags': {'count': 0}}, False),
        ]
        for rule, names, expected in rules:
            audience = Audience(
                rule,
                allocations=[
                    Allocation('chartreuse', percent=100),
                ],
            ).validate()
            self.assertEqual(audience.matches(names), expected)

    def test_determine_variant(self):
        audience = Audience(
            None,
            allocations=[
                Allocation('red', percent=25),
                Allocation('green', percent=25),
                Allocation('blue', percent=25),
                Allocation('yellow', percent=25),
            ],
        ).validate()
        cases = [
            # Test key points at edges and between ranges.
            [0, 'red'],
            [1, 'red'],
            [24, 'red'],
            [25, 'green'],
            [37, 'green'],
            [49, 'green'],
            [50, 'blue'],
            [65, 'blue'],
            [74, 'blue'],
            [75, 'yellow'],
            [90, 'yellow'],
            [99, 'yellow'],
        ]
        for rangekey, expected_variant in cases:
            self.assertEqual(
                audience.determine_variant(rangekey), expected_variant
            )


class TestExperiment(SimpleTestCase):
    def test_seed_initialized_to_test_name_if_not_provided(self):
        name = 'experiment_mc_fancy_test'
        exp = Experiment(
            name=name,
            identity='tk',
            variants=['exuberant'],
            audiences=[
                Audience(None, [Allocation('exuberant', 100)]),
            ],
        ).validate()
        self.assertEqual(exp.seed, name)

    def test_validate_invalid_experiment_name_raises(self):
        for name in [*r' ~!@#$%^&*()`-+=[]{};:<>,./\?"', "'", '']:
            with self.assertRaisesRegex(ValueError, 'Name must contain only'):
                Experiment(
                    name=name,
                    identity='tk',
                    variants=['red'],
                    audiences=[
                        Audience(None, [Allocation('red', 100)]),
                    ],
                ).validate()

    def test_validate_no_variants(self):
        with self.assertRaisesRegex(
            ValueError, 'must define at least one variant'
        ):
            Experiment(
                name='experiment',
                identity='tk',
                variants=[],
                audiences=[
                    Audience(None, [Allocation('red', 100)]),
                ],
            ).validate()

    def test_validate_identity_random(self):
        Experiment(
            name='experiment',
            identity='random',
            variants=['red'],
            audiences=[
                Audience(None, [Allocation('red', 100)]),
            ],
        ).validate()

    def test_validate_invalid_identity_name(self):
        with self.assertRaisesRegex(ValueError, 'Invalid identity symbol name'):
            Experiment(
                name='experiment',
                identity='jkj_*33',
                variants=['gold'],
                audiences=[
                    Audience(None, [Allocation('gold', 100)]),
                ],
            ).validate()

    def test_validate_no_audiences(self):
        with self.assertRaisesRegex(ValueError, 'at least one audience'):
            Experiment(
                name='experiment',
                identity='random',
                variants=['silver'],
                audiences=[],
            ).validate()

    def test_validate_last_audience_rule_not_none(self):
        with self.assertRaisesRegex(
            ValueError, 'Last audience must not define a rule'
        ):
            Experiment(
                name='experiment',
                identity='random',
                variants=['silver'],
                audiences=[
                    Audience('user.id == 1', [Allocation('gold', 100)]),
                ],
            ).validate()

    def test_validate_multiple_none_rules(self):
        with self.assertRaisesRegex(
            ValueError, 'Only the last audience rule can be None'
        ):
            Experiment(
                name='experiment',
                identity='random',
                variants=['silver', 'gold'],
                audiences=[
                    Audience(None, [Allocation('silver', 100)]),
                    Audience(None, [Allocation('gold', 100)]),
                ],
            ).validate()

    def test_validate_audience_with_undeclared_variant(self):
        with self.assertRaisesRegex(ValueError, 'Undeclared variant'):
            Experiment(
                name='experiment',
                identity='random',
                variants=['gold'],
                audiences=[
                    Audience('user.id == 1', [Allocation('gold', 100)]),
                    Audience(None, [Allocation('ruby', 100)]),
                ],
            ).validate()

    def test_determine_variant(self):
        exp = Experiment(
            name='experiment',
            identity='random',
            variants=['gold', 'ruby'],
            audiences=[
                Audience('user.id == 1', [Allocation('gold', 100)]),
                Audience(None, [Allocation('ruby', 100)]),
            ],
        ).validate()
        context_ruby = Context({'user': User(id=0)})
        context_gold = Context({'user': User(id=1)})
        self.assertEqual(exp.determine_variant(context_ruby, None), 'ruby')
        self.assertEqual(exp.determine_variant(context_gold, None), 'gold')

    def test_determine_variant_override(self):
        exp = Experiment(
            name='experiment',
            identity='random',
            variants=['gold', 'ruby'],
            audiences=[
                Audience('user.id == 1', [Allocation('gold', 100)]),
                Audience(None, [Allocation('ruby', 100)]),
            ],
        ).validate()
        context_ruby = Context({'user': User(id=0)})
        context_gold = Context({'user': User(id=1)})
        self.assertEqual(exp.determine_variant(context_ruby, 'gold'), 'gold')
        self.assertEqual(exp.determine_variant(context_gold, 'ruby'), 'ruby')

    def test_determine_variant_ignores_unknown_variant_override(self):
        exp = Experiment(
            name='experiment',
            identity='random',
            variants=['gold', 'ruby'],
            audiences=[
                Audience('user.id == 1', [Allocation('gold', 100)]),
                Audience(None, [Allocation('ruby', 100)]),
            ],
        ).validate()
        context_ruby = Context({'user': User(id=0)})
        self.assertEqual(exp.determine_variant(context_ruby, 'wumbo'), 'ruby')


class TestExtractOverrides(SimpleTestCase):
    factory = RequestFactory()

    def test_no_overrides(self):
        request = self.factory.get('/')

        overrides = _extract_overrides(request)
        self.assertEqual(overrides, {})

    def test_extracts_overrides(self):
        request = self.factory.get('/')
        request.COOKIES[DEFAULT_CRAVENSWORTH_COOKIE] = 'switch:on'

        overrides = _extract_overrides(request)
        self.assertEqual(overrides, {'switch': 'on'})

    def test_multiple_overrides(self):
        request = self.factory.get('/')
        request.COOKIES[DEFAULT_CRAVENSWORTH_COOKIE] = (
            'experiment1:variant1 experiment2:variant2'
        )

        overrides = _extract_overrides(request)
        self.assertEqual(
            overrides,
            {
                'experiment1': 'variant1',
                'experiment2': 'variant2',
            },
        )

    def test_duplicate_overrides(self):
        request = self.factory.get('/')
        request.COOKIES[DEFAULT_CRAVENSWORTH_COOKIE] = (
            'experiment:variant1 experiment:variant2'
        )

        overrides = _extract_overrides(request)
        # We don't really care about duplicates. It'll be sorted out by the
        # state, where last one clobbers all.
        self.assertEqual(
            overrides,
            {
                'experiment': 'variant2',
            },
        )

    @override_settings(CRAVENSWORTH={'OVERRIDE_COOKIE': 'mycookie'})
    def test_custom_cookie_name(self):
        request = self.factory.get('/')
        request.COOKIES['mycookie'] = 'swank:active'

        overrides = _extract_overrides(request)
        self.assertEqual(overrides, {'swank': 'active'})

    @override_settings(CRAVENSWORTH={'ENABLED_IPS': []})
    def test_ip_restricted_no_ips(self):
        request = self.factory.get('/')
        request.COOKIES[DEFAULT_CRAVENSWORTH_COOKIE] = 'switch:on'

        overrides = _extract_overrides(request)
        self.assertEqual(overrides, {})

    @override_settings(CRAVENSWORTH={'ENABLED_IPS': ['127.0.0.1']})
    def test_ip_restricted_matching_ip(self):
        request = self.factory.get('/')
        request.META['REMOTE_ADDR'] = '127.0.0.1'
        request.COOKIES[DEFAULT_CRAVENSWORTH_COOKIE] = 'switch:on'

        overrides = _extract_overrides(request)
        self.assertEqual(overrides, {'switch': 'on'})

    @override_settings(CRAVENSWORTH={'ENABLED_IPS': ['127.0.0.1', '127.0.0.2']})
    def test_ip_restricted_unknown_ip(self):
        request = self.factory.get('/')
        request.META['REMOTE_ADDR'] = '127.0.0.3'
        request.COOKIES[DEFAULT_CRAVENSWORTH_COOKIE] = 'switch:on'

        overrides = _extract_overrides(request)
        self.assertEqual(overrides, {})


class TestCravensworthState(SimpleTestCase):
    def test_returns_false_for_undeclared_experiment(self):
        state = _CravensworthState(
            experiments=[],
            overrides={},
            context=Context(),
        )
        self.assertFalse(state.is_variant('bobobobobobobo', 'bam'))

    def test_is_variant_single_variant(self):
        state = _CravensworthState(
            experiments=[
                Experiment(
                    name='a',
                    identity='random',
                    variants=['1', '2'],
                    audiences=[
                        Audience(rule=None, allocations=[Allocation('1', 100)])
                    ],
                )
            ],
            overrides={},
            context=Context(),
        )
        self.assertTrue(state.is_variant('a', '1'))
        self.assertFalse(state.is_variant('a', '2'))

    def test_is_variant_multiple_variants(self):
        state = _CravensworthState(
            experiments=[
                Experiment(
                    name='a',
                    identity='random',
                    variants=['1', '2', '3'],
                    audiences=[
                        Audience(rule=None, allocations=[Allocation('1', 100)])
                    ],
                )
            ],
            overrides={},
            context=Context(),
        )
        self.assertTrue(state.is_variant('a', ['1', '2']))
        self.assertFalse(state.is_variant('a', ['2', '3']))

    def test_override(self):
        state = _CravensworthState(
            experiments=[
                Experiment(
                    name='a',
                    identity='random',
                    variants=['1', '2'],
                    audiences=[
                        Audience(rule=None, allocations=[Allocation('1', 100)])
                    ],
                )
            ],
            overrides={'a': '2'},
            context=Context(),
        )
        self.assertFalse(state.is_variant('a', '1'))
        self.assertTrue(state.is_variant('a', '2'))



from django.http import HttpResponse, Http404
from django.test import RequestFactory, SimpleTestCase

from cravensworth.core.models import (
    Allocation,
    Audience,
    Context,
    Experiment,
)
from cravensworth.core.experiment import _CravensworthState, set_state
from cravensworth.core.decorators import variant

from tests.testapp.models import FancyModel


class TestVariant(SimpleTestCase):
    def setUp(self):
        factory = RequestFactory()
        state = _CravensworthState(
            experiments=[
                Experiment(
                    name='on_switch',
                    identity='random',
                    variants=('on', 'off'),
                    audiences=(
                        Audience(
                            rule=None,
                            allocations=(
                                Allocation(variant='on', percent=100),
                            ),
                        ),
                    ),
                ).validate(),
                Experiment(
                    name='off_switch',
                    identity='random',
                    variants=('on', 'off'),
                    audiences=(
                        Audience(
                            rule=None,
                            allocations=(
                                Allocation(variant='off', percent=100),
                            ),
                        ),
                    ),
                ).validate(),
                Experiment(
                    name='multivariant',
                    identity='random',
                    variants=('active', 'inactive', 'control'),
                    audiences=(
                        Audience(
                            rule=None,
                            allocations=(
                                Allocation(variant='active', percent=100),
                            ),
                        ),
                    ),
                ).validate(),
            ],
            overrides={},
            context=Context(),
        )
        self.request = factory.get('/')
        set_state(self.request, state)

    def test_renders_view_when_variant_matches(self):
        @variant('on_switch', 'on')
        def some_page(request):
            return HttpResponse('OK')

        response = some_page(self.request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'OK')

    def test_renders_view_when_variant_list_matches(self):
        @variant('multivariant', ['active', 'control'])
        def some_page(request):
            return HttpResponse('OK')

        response = some_page(self.request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'OK')

    def test_redirects_to_model_view_when_variant_does_not_match(self):
        @variant('off_switch', 'on', redirect_to=FancyModel)
        def some_page(request):
            return HttpResponse('OK')

        response = some_page(self.request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers.get('location'), '/model-redirected/')

    def test_redirects_to_model_view_when_variant_list_does_not_match(self):
        @variant(
            'multivariant', ['control', 'inactive'], redirect_to=FancyModel
        )
        def some_page(request):
            return HttpResponse('OK')

        response = some_page(self.request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers.get('location'), '/model-redirected/')

    def test_redirects_to_named_view_when_variant_does_not_match(self):
        @variant('off_switch', 'on', redirect_to='view-redirect')
        def some_page(request):
            return HttpResponse('OK')

        response = some_page(self.request)
        assert response.status_code == 302
        assert response.headers.get('location') == '/redirected/'

    def test_redirects_to_url_when_variant_does_not_match(self):
        @variant('off_switch', 'on', redirect_to='https://www.example.com/')
        def some_page(request):
            return HttpResponse('OK')

        response = some_page(self.request)
        assert response.status_code == 302
        assert response.headers.get('location') == 'https://www.example.com/'

    def test_raises_404_when_no_redirect_provided(self):
        @variant('off_switch', 'on')
        def some_page(request):
            return HttpResponse('OK')

        with self.assertRaises(Http404):
            some_page(self.request)



from django.urls import path, include

from .views import (
    fancy_model_view,
    named_redirect_view,
    on_page,
    off_page,
    home,
    switchon_single_view,
    switchon_double_view,
    switchon_variable_view,
    switchon_content_view,
    switchoff_single_view,
    switchoff_double_view,
    switchoff_variable_view,
    switchoff_content_view,
    variant_single_view,
    variant_multiple_view,
    variant_else_view,
    variant_none_view,
    variant_unknown_view,
    variant_variable_view,
)

template_tag_test_urls = [
    path('/', home, name='home'),
    # Switchon
    path('on-single/', switchon_single_view),
    path('on-double/', switchon_double_view),
    path('on-variable/', switchon_variable_view),
    path('on-content/', switchon_content_view),
    # Switchoff
    path('off-single/', switchoff_single_view),
    path('off-double/', switchoff_double_view),
    path('off-variable/', switchoff_variable_view),
    path('off-content/', switchoff_content_view),
    path('variant-single/', variant_single_view),
    path('variant-multiple/', variant_multiple_view),
    path('variant-else/', variant_else_view),
    path('variant-none/', variant_none_view),
    path('variant-unknown/', variant_unknown_view),
    path('variant-variable/', variant_variable_view),
]

urlpatterns = [
    path('redirected/', named_redirect_view, name='view-redirect'),
    path('model-redirected/', fancy_model_view, name='model-view-redirect'),
    path('active/', on_page),
    path('inactive/', off_page),
    path('templates/', include(template_tag_test_urls)),
]



from cravensworth.core.experiment import get_state
from cravensworth.core.utils import set_tracking_key


def cravensworth_middleware(get_response):
    """
    Middleware that extracts overrides from incoming requests and eagerly adds
    cravensworth data to the request.

    This should be added to the `MIDDLEWARE` list in your Django settings.
    """

    def middleware(request):
        get_state(request)
        response = get_response(request)
        set_tracking_key(request, response)
        return response

    return middleware



from django.test import RequestFactory, SimpleTestCase, override_settings

from testapp import views


class TestVariant(SimpleTestCase):
    factory = RequestFactory()

    @override_settings(
        CRAVENSWORTH={
            'EXPERIMENTS': [
                {
                    'name': 'exp_1',
                    'identity': 'random',
                    'variants': [
                        {'name': 'active'},
                        {'name': 'inactive'},
                        {'name': 'control'},
                    ],
                    'audiences': [
                        {
                            'rule': None,
                            'allocations': [
                                {'variant': 'active', 'percent': 100},
                            ],
                        },
                    ],
                }
            ]
        }
    )
    def test_single_variant(self):
        request = self.factory.get('/templates/variant-single/')
        response = views.variant_single_view(request)
        self.assertContains(response, ':ACTIVE:')

    @override_settings(
        CRAVENSWORTH={
            'EXPERIMENTS': [
                {
                    'name': 'exp_1',
                    'identity': 'random',
                    'variants': [
                        {'name': 'active'},
                        {'name': 'inactive'},
                        {'name': 'control'},
                    ],
                    'audiences': [
                        {
                            'rule': None,
                            'allocations': [
                                {'variant': 'active', 'percent': 100},
                            ],
                        },
                    ],
                }
            ]
        }
    )
    def test_variant_else(self):
        request = self.factory.get('/templates/variant-else/')
        response = views.variant_else_view(request)
        self.assertContains(response, ':ELSE:')

    @override_settings(
        CRAVENSWORTH={
            'EXPERIMENTS': [
                {
                    'name': 'exp_1',
                    'identity': 'random',
                    'variants': [
                        {'name': 'one'},
                        {'name': 'two'},
                        {'name': 'three'},
                        {'name': 'four'},
                        {'name': 'five'},
                    ],
                    'audiences': [
                        {
                            'rule': None,
                            'allocations': [
                                {'variant': 'four', 'percent': 100},
                            ],
                        },
                    ],
                }
            ]
        }
    )
    def test_multiple_variants(self):
        request = self.factory.get('/templates/variant-multiple/')
        response = views.variant_multiple_view(request)
        self.assertContains(response, ':THREE-FOUR:')

    @override_settings(
        CRAVENSWORTH={
            'EXPERIMENTS': [
                {
                    'name': 'exp_1',
                    'identity': 'random',
                    'variants': [
                        {'name': 'active'},
                        {'name': 'inactive'},
                    ],
                    'audiences': [
                        {
                            'rule': None,
                            'allocations': [
                                {'variant': 'inactive', 'percent': 100},
                            ],
                        },
                    ],
                }
            ]
        }
    )
    def test_no_match(self):
        request = self.factory.get('/templates/variant-none/')
        response = views.variant_none_view(request)
        self.assertNotContains(response, ':ACTIVE:')

    @override_settings(
        CRAVENSWORTH={
            'EXPERIMENTS': [
                {
                    'name': 'exp_1',
                    'identity': 'random',
                    'variants': [
                        {'name': 'active'},
                        {'name': 'inactive'},
                    ],
                    'audiences': [
                        {
                            'rule': None,
                            'allocations': [
                                {'variant': 'active', 'percent': 100},
                            ],
                        },
                    ],
                }
            ]
        }
    )
    def test_variables(self):
        request = self.factory.get('/templates/variant-variable/')
        response = views.variant_variable_view(request)
        self.assertContains(response, ':ACTIVE:')

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': []})
    def test_unknown_experiment(self):
        request = self.factory.get('/templates/variant-unknown/')
        response = views.variant_unknown_view(request)
        self.assertNotContains(response, ':UNKNOWN:')
        self.assertContains(response, ':ELSE:')


class TestSwitchOn(SimpleTestCase):
    factory = RequestFactory()

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:on']})
    def test_switch_on_single(self):
        request = self.factory.get('/templates/on-single/')
        response = views.switchon_single_view(request)
        self.assertContains(response, 'ON')

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:off']})
    def test_switch_off_single(self):
        request = self.factory.get('/templates/on-single/')
        response = views.switchon_single_view(request)
        self.assertNotContains(response, 'ON')

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:on']})
    def test_switch_on_double(self):
        request = self.factory.get('/templates/on-double/')
        response = views.switchon_double_view(request)
        self.assertContains(response, 'ON')
        self.assertNotContains(response, 'OFF')

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:off']})
    def test_switch_off_double(self):
        request = self.factory.get('/templates/on-double/')
        response = views.switchon_double_view(request)
        self.assertNotContains(response, 'ON')
        self.assertContains(response, 'OFF')

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:on']})
    def test_switch_on_variable(self):
        request = self.factory.get('/templates/on-variable/')
        response = views.switchon_variable_view(request)
        self.assertContains(response, 'ON')

    @override_settings(
        CRAVENSWORTH={
            'SWITCHES': [
                {'name': 'active', 'on': True},
                {'name': 'inactive', 'on': False},
            ],
        }
    )
    @override_settings(
        CRAVENSWORTH={
            'EXPERIMENTS': ['active:on', 'inactive:off'],
        }
    )
    def test_switch_template_content(self):
        request = self.factory.get('/templates/on-content/')
        response = views.switchon_content_view(request)
        self.assertContains(response, ':ACTIVE-ON:')
        self.assertContains(response, ':INACTIVE-OFF:')
        self.assertNotContains(response, ':INACTIVE-ON:')


class TestSwitchOff(SimpleTestCase):
    factory = RequestFactory()

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:off']})
    def test_switch_off_single(self):
        request = self.factory.get('/templates/off-single/')
        response = views.switchoff_single_view(request)
        self.assertContains(response, 'OFF')

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:on']})
    def test_switch_on_single(self):
        request = self.factory.get('/templates/off-single/')
        response = views.switchoff_single_view(request)
        self.assertNotContains(response, 'OFF')

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:off']})
    def test_switch_off_double(self):
        request = self.factory.get('/templates/off-double/')
        response = views.switchoff_double_view(request)
        self.assertContains(response, 'OFF')
        self.assertNotContains(response, 'ON')

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:on']})
    def test_switch_on_double(self):
        request = self.factory.get('/templates/off-double/')
        response = views.switchoff_double_view(request)
        self.assertNotContains(response, 'OFF')
        self.assertContains(response, 'ON')

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:off']})
    def test_switch_off_variable(self):
        request = self.factory.get('/templates/off-variable/')
        response = views.switchoff_variable_view(request)
        self.assertContains(response, 'OFF')

    @override_settings(
        CRAVENSWORTH={
            'EXPERIMENTS': ['active:on', 'inactive:off'],
        }
    )
    def test_switch_template_content(self):
        request = self.factory.get('/templates/off-content/')
        response = views.switchoff_content_view(request)
        self.assertNotContains(response, ':ACTIVE-OFF:')
        self.assertContains(response, ':INACTIVE-OFF:')
        self.assertNotContains(response, 'INACTIVE-ON')



from datetime import timedelta
from unittest.mock import patch

from django.contrib.auth.models import AnonymousUser
from django.http import HttpResponse
from django.test import Client, RequestFactory, TestCase, override_settings

from cravensworth.core.middleware import cravensworth_middleware


@override_settings(CRAVENSWORTH={'TRACKING_COOKIE': 'tk'})
class TestCravensworthMiddleware(TestCase):
    client = Client()
    factory = RequestFactory()

    def test_sets_tracking_key(self):
        with patch(
            'cravensworth.core.utils.generate_tracking_key'
        ) as mock_generate_tracking_key:
            mock_generate_tracking_key.return_value = 'NEW_TRACKING_KEY'
            response = self.client.get('/')
        self.assertTrue('tk' in response.cookies)

        cookie = response.cookies['tk']
        self.assertEqual(cookie.value, 'NEW_TRACKING_KEY')
        self.assertEqual(cookie['max-age'], timedelta(days=365).total_seconds())
        self.assertEqual(cookie['path'], '/')
        self.assertEqual(cookie['httponly'], True)
        self.assertEqual(cookie['samesite'], 'Lax')

    def test_uses_existing_tracking_key(self):
        self.client.cookies.load({'tk': 'THE_TRACKING_KEY'})
        response = self.client.get('/')
        self.assertTrue('tk' in response.cookies)
        self.assertEqual(response.cookies['tk'].value, 'THE_TRACKING_KEY')

    def test_sets_state(self):
        middleware = cravensworth_middleware(lambda r: HttpResponse('OK'))
        request = self.factory.get('/')
        request.user = AnonymousUser()
        middleware(request)
        state = request._cravensworth_state
        self.assertIsNotNone(state)



from django import template
from django.template import Node, TemplateSyntaxError

from cravensworth.core.experiment import is_on, is_variant


register = template.Library()


class _VariantBranch:
    """
    Represents a single variant or elif branch within a variant template tag.

    This class holds the experiment name, the specific variants associated with
    this branch, and the nodelist to render if this branch is active.
    """

    def __init__(self, experiment_name_var, variant_vars, nodelist):
        self.experiment_name_var = experiment_name_var
        self.variant_vars = variant_vars
        self.nodelist = nodelist

    def is_active(self, context, request, experiment_name) -> bool:
        variants = []
        for variant_var in self.variant_vars:
            variant = variant_var.resolve(context)
            if isinstance(variant, (list, tuple)):
                variants.extend(variant)
            else:
                variants.append(variant)

        return is_variant(request, experiment_name, variants)

    def render(self, context):
        return self.nodelist.render(context)


class _VariantBlockNode(Node):
    """
    Represents a `variant` template tag and its various possible branches,
    similar to an if/elif/else block.
    """

    def __init__(self, experiment_name_var, branches, else_nodelist):
        self.experiment_name_var = experiment_name_var
        self.branches = branches
        self.else_nodelist = else_nodelist

    def render(self, context):
        request = context['request']
        experiment_name = self.experiment_name_var.resolve(context)

        for branch in self.branches:
            if branch.is_active(context, request, experiment_name):
                return branch.render(context)

        if self.else_nodelist:
            return self.else_nodelist.render(context)
        return ''


@register.tag('variant')
def variant(parser, token):
    """
    Renders content based on active experiment variants, similar to if/elif/
    else.

    `variant` and `elif` blocks can take one or more variant. If no named
    variants are active the else block, if provided, will be rendered.

    Example:

        {% variant "experiment_name" "variant_A" "variant_B" %}
            Content for variant A or B...
        {% elif "variant_C" %}
            Optional Content for variant C...
        {% else %}
            Optional fallback content...
        {% endvariant %}
    """
    args = token.split_contents()
    tag_name = args.pop(0)

    if len(args) < 2:
        raise TemplateSyntaxError(
            f"'{tag_name}' requires at least an experiment name and variant"
        )

    experiment_name_var = template.Variable(args.pop(0))
    variant_vars = [template.Variable(arg) for arg in args]
    branches = []

    # Parse main block
    nodelist = parser.parse(('elif', 'else', 'endvariant'))
    branches.append(_VariantBranch(experiment_name_var, variant_vars, nodelist))

    # Parse optional elif/else blocks.
    else_nodelist = None

    while True:
        token = parser.next_token()

        if token.contents.startswith('elif'):
            args = token.split_contents()
            args.pop(0)  # Skip 'elif'
            if not args:
                raise TemplateSyntaxError(
                    "'elif' requires at least one variant"
                )
            variant_vars = [template.Variable(b) for b in args]
            nodelist = parser.parse(('elif', 'else', 'endvariant'))
            branches.append(
                _VariantBranch(experiment_name_var, variant_vars, nodelist)
            )

        elif token.contents == 'else':
            else_nodelist = parser.parse(('endvariant',))
            parser.delete_first_token()
            break

        elif token.contents == 'endvariant':
            break

        else:
            raise TemplateSyntaxError(f'Unexpected tag: {token.contents}')

    return _VariantBlockNode(experiment_name_var, branches, else_nodelist)


class _SwitchNode(Node):
    def __init__(self, name, nodelist_on, nodelist_off):
        self.name = template.Variable(name)
        self.nodelist_on = nodelist_on
        self.nodelist_off = nodelist_off

    def render(self, context):
        request = context['request']
        name = self.name.resolve(context)
        nodelist = (
            self.nodelist_on if is_on(request, name) else self.nodelist_off
        )
        return nodelist.render(context) if nodelist is not None else ''


def _parse_switch_node(parser, token):
    try:
        tag_name, switch_name = token.split_contents()
    except ValueError:
        raise TemplateSyntaxError(f'{tag_name} tag requires a switch name')

    end_tag_name = f'end{tag_name}'
    nodelist_first = parser.parse(('else', end_tag_name))
    token = parser.next_token()

    if token.contents == 'else':
        nodelist_second = parser.parse((end_tag_name,))
        parser.delete_first_token()
    else:
        nodelist_second = None

    return switch_name, nodelist_first, nodelist_second


@register.tag(name='switchon')
def switchon(parser, token):
    """
    Template tag that conditionally renders a block if a named switch is on.

    `switchon` is like an if/else that renders content conditionally based on
    the active variant of an experiment. switchon is similar to the `variant`
    tag, but with the implicit variant 'on' for the primary block and 'off' for
    the else block.

    To use `switchon` with an experiment, it should be configured as a switch:
        declare only `on` or `off`
    as variants.

    Example:

        {% switchon "super_cool_feature" %}
            Content displayed if the switch is on
        {% else %}
            Optional content displayed if the switch is off...
        {% endswitchon %}
    """
    switch_name, nodelist_on, nodelist_off = _parse_switch_node(parser, token)
    return _SwitchNode(switch_name, nodelist_on, nodelist_off)


@register.tag(name='switchoff')
def switchoff(parser, token):
    """
    Template tag that conditionally renders a block if a named switch is off.

    switchoff is like an if/else that renders content conditionally based on the
    active variant of an experiment. switchoff is similar to the `variant` tag,
    but with the implicit variant 'off' for the primary block and 'on' for the
    else block.

    To use `switchoff` with an experiment, it should declare only `on` or `off`
    as variants.

    Example:

        {% switchoff "super_cool_feature" %}
            Content displayed if the switch is off
        {% else %}
            Optional content displayed if the switch is on...
        {% endswitchoff %}
    """
    switch_name, nodelist_off, nodelist_on = _parse_switch_node(parser, token)
    return _SwitchNode(switch_name, nodelist_on, nodelist_off)


__all__ = [
    'variant',
    'switchon',
    'switchoff',
]



from django.http import HttpResponse
from django.views.generic import TemplateView

from cravensworth.core.decorators import switch_on


def home(request):
    return HttpResponse('OK')


def fancy_model_view(request):
    return HttpResponse('OK')


def named_redirect_view(request):
    return HttpResponse('OK')


@switch_on('active')
def on_page(request):
    return HttpResponse('Active page')


@switch_on('inactive')
def off_page(request):
    return HttpResponse('Inactive page')


# Template tag test views
switchon_single_view = TemplateView.as_view(
    template_name='switchon_single.html'
)
switchon_double_view = TemplateView.as_view(
    template_name='switchon_double.html'
)
switchon_variable_view = TemplateView.as_view(
    template_name='switchon_variable.html'
)
switchon_content_view = TemplateView.as_view(
    template_name='switchon_content.html'
)

switchoff_single_view = TemplateView.as_view(
    template_name='switchoff_single.html'
)
switchoff_double_view = TemplateView.as_view(
    template_name='switchoff_double.html'
)
switchoff_variable_view = TemplateView.as_view(
    template_name='switchoff_variable.html'
)
switchoff_content_view = TemplateView.as_view(
    template_name='switchoff_content.html'
)

variant_single_view = TemplateView.as_view(template_name='variant_single.html')
variant_multiple_view = TemplateView.as_view(
    template_name='variant_multiple.html'
)
variant_else_view = TemplateView.as_view(template_name='variant_else.html')
variant_none_view = TemplateView.as_view(template_name='variant_none.html')
variant_unknown_view = TemplateView.as_view(
    template_name='variant_unknown.html'
)
variant_variable_view = TemplateView.as_view(
    template_name='variant_variable.html'
)



from datetime import timedelta
from uuid import uuid4
from django.http import HttpRequest, HttpResponse

from cravensworth.core.conf import get_setting


DEFAULT_TRACKING_COOKIE = '__cwtk'


def _tracking_cookie() -> str:
    return get_setting('TRACKING_COOKIE', DEFAULT_TRACKING_COOKIE)


def generate_tracking_key() -> str:
    """
    Generate a new tracking key.
    """
    return uuid4().hex


def get_tracking_key(request: HttpRequest) -> str:
    """
    Gets the tracking key from a request if it exists. If the tracking key does
    not exist, one will be created and added to the request, so it may be
    consumed later, if required.
    """
    tk = getattr(request, '_cravensworth_tk', None)
    if tk is None:
        tk = request.COOKIES.get(_tracking_cookie())
    if tk is None:
        tk = generate_tracking_key()
    setattr(request, '_cravensworth_tk', tk)
    return tk


def set_tracking_key(request: HttpRequest, response: HttpResponse):
    """
    Sets the current tracking key as a cookie on the provided response. If no
    key exists, a new one will be generated.
    """
    name = _tracking_cookie()
    value = get_tracking_key(request)
    response.set_cookie(
        name,
        value,
        max_age=timedelta(days=365),
        httponly=True,
        samesite='Lax',
    )


__all__ = []



from typing import overload

from django.http import HttpRequest

from cravensworth.core.conf import get_setting
from cravensworth.core.experiment import DEFAULT_CRAVENSWORTH_COOKIE


@overload
def override_experiment(request: HttpRequest, experiment: dict[str, str]): ...


@overload
def override_experiment(
    request: HttpRequest, experiment: str, variant: str
): ...


def override_experiment(
    request: HttpRequest,
    experiment: dict[str, str] | str,
    variant: str | None = None,
):
    """Override experiment variants on a request by setting the override cookie.

    This is useful in tests to make experiment behavior deterministic.

    Can be called with a single experiment name and variant::

        override_experiment(request, 'switch', 'on')

    Or with a dict to override multiple experiments at once::

        override_experiment(request, {'switch1': 'on', 'switch2': 'off'})

    Args:
        request: The Django request object to set the override cookie on.
        experiment: Either an experiment name (str) or a dict mapping
            experiment names to variant values.
        variant: The variant value when ``experiment`` is a str.
            Required when ``experiment`` is a str, ignored otherwise.

    Raises:
        ValueError: If ``experiment`` is a str and ``variant`` is None.
        TypeError: If ``experiment`` is neither a str nor a dict.
    """

    def make_override(name, value):
        return f'{name}:{value}'

    if isinstance(experiment, str):
        if variant is None:
            raise ValueError('variant must not be None')
        cookie = make_override(experiment, variant)

    elif isinstance(experiment, dict):
        cookie = ' '.join(
            make_override(name, value) for name, value in experiment.items()
        )

    else:
        raise TypeError('expected a string or a dictionary for experiment')

    cookie_name = get_setting('OVERRIDE_COOKIE', DEFAULT_CRAVENSWORTH_COOKIE)
    request.COOKIES[cookie_name] = cookie



from typing import overload, Any
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


_NotGiven = object()


@overload
def get_setting(path: str, default: Any) -> Any: ...
@overload
def get_setting(path: str) -> Any: ...


def get_setting(path: str, default: Any = _NotGiven):
    """
    Retrieves a cravensworth configuration setting.

    This is not for general Django settings. To get Django settings, use the
    settings object from django.conf directly.

    Args:
        path (str): The dotted path to the setting.
        default (Any, optional): A fallback value to return if the setting is
            not found. If not default is provided, an error will be raised.

    Returns:
        Any: The setting value, or `default` if not found and provided.

    Raises:
        ImproperlyConfigured: The setting is not found and no default is given.
    """
    cursor = getattr(settings, 'CRAVENSWORTH')
    for segment in path.split('.'):
        if segment in cursor:
            cursor = cursor.get(segment)
        else:
            if default == _NotGiven:
                raise ImproperlyConfigured(
                    f'Expected setting CRAVENSWORTH.{path} not found',
                )
            return default
    return cursor



import logging
from typing import Iterable

from django.http import HttpRequest

from cravensworth.core.conf import get_setting
from cravensworth.core.models import Context, Experiment
from cravensworth.core.providers import get_context_provider, get_source


logger = logging.getLogger(__name__)


class _CravensworthState:
    """
    A container for holding the experiment state for a particular entity within
    a given lifetime (e.g., a single request).
    """

    def __init__(
        self,
        experiments: Iterable[Experiment],
        overrides: dict[str, str],
        context: Context,
    ):
        self._experiments = {e.name: e for e in experiments}
        self._overrides = overrides
        self._context = context

    def is_variant(self, name: str, variant: str | list[str]) -> bool:
        """
        Returns true if the determined variant for the current entity matches
        `variant` or one of the list of variants, if multiple.
        """
        experiment = self._experiments.get(name)
        if experiment is None:
            logger.warning(
                'is_variant was called for an undeclared experiment "%s". If '
                'this is a valid experiment, ensure that it exists in your'
                'experiment source and is correctly configured. Returning non-'
                'match.',
                name,
            )
            return False

        override = self._overrides.get(experiment.name)
        active_variant = experiment.determine_variant(self._context, override)
        return active_variant in (
            variant if isinstance(variant, list) else [variant]
        )

    def export(self) -> dict[str, str]:
        state = {}
        for experiment in self._experiments.values():
            override = self._overrides[experiment.name]
            state[experiment.name] = experiment.determine_variant(
                self._context,
                override,
            )
        return state


DEFAULT_CRAVENSWORTH_COOKIE = '__cw'


def _extract_overrides(request: HttpRequest) -> dict[str, str]:
    """
    Extracts experiment overrides from the given request and returns them as
    a mapping of experiment to overridden variant.

    This method takes into account whether the application has configured IP
    restriction. If IP restriction is enabled and the IP address does not match
    the list of allowed IPs, overrides will be empty.
    """
    enabled_ips = get_setting('ENABLED_IPS', None)
    restrict_ips = enabled_ips is not None
    overrides = {}

    if not restrict_ips or request.META['REMOTE_ADDR'] in enabled_ips:
        cookie_name = get_setting(
            'OVERRIDE_COOKIE', DEFAULT_CRAVENSWORTH_COOKIE
        )
        cookie = request.COOKIES.get(cookie_name)
        if cookie is not None:
            for override in cookie.split():
                experiment, variant = override.rsplit(':', maxsplit=1)
                overrides[experiment] = variant

    return overrides


class _StateResolver:
    def __init__(self):
        self.source = get_source()
        self.context_provider = get_context_provider()

    def resolve(self, request):
        experiments = self.source.load()
        overrides = _extract_overrides(request)
        context = self.context_provider.context(request=request)
        return _CravensworthState(experiments, overrides or {}, context)


_state_resolver = _StateResolver()


def set_state(request: HttpRequest, state: _CravensworthState):
    """
    Sets experiment state on the given request.
    """
    setattr(request, '_cravensworth_state', state)


def get_state(request: HttpRequest) -> _CravensworthState:
    """
    Gets experiment state from the given request.
    """
    if not hasattr(request, '_cravensworth_state'):
        set_state(request, _state_resolver.resolve(request))
    return getattr(request, '_cravensworth_state', None)


def is_variant(
    request: HttpRequest, name: str, variant: str | list[str]
) -> bool:
    """
    Returns true if the determined variant for the current entity matches
    `variant` or one of the list of variants, if multiple.
    """
    state = get_state(request)
    return state.is_variant(name, variant)


def is_on(request: HttpRequest, name: str) -> bool:
    """
    Returns True if the named switch is on; false otherwise.
    """
    return is_variant(request, name, 'on')


def is_off(request: HttpRequest, name: str) -> bool:
    """
    Returns True if the named switch is off; false otherwise.
    """
    return is_variant(request, name, 'off')


__all__ = [
    'is_variant',
    'is_on',
    'is_off',
]



from django.test import TestCase, override_settings

from cravensworth.core import models
from cravensworth.core.providers import SettingsSource


class TestSettingsSource(TestCase):
    source = SettingsSource()

    @override_settings(
        CRAVENSWORTH={
            'EXPERIMENTS': [
                'switch_syntax:on',
                {
                    'name': 'experimentitious',
                    'identity': 'user.id',
                    'seed': 'maseed',
                    'variants': [
                        {'name': 'active'},
                        {'name': 'inactive'},
                        {'name': 'control'},
                    ],
                    'audiences': [
                        {
                            'rule': 'locale == "en-US"',
                            'allocations': [
                                {'variant': 'active', 'percent': 10},
                                {'variant': 'inactive', 'percent': 80},
                                {'variant': 'control', 'percent': 10},
                            ],
                        },
                        {
                            'allocations': [
                                {'variant': 'inactive', 'percent': 100},
                            ]
                        },
                    ],
                },
            ],
        }
    )
    def test_experiment_construction(self):
        experiments = self.source.load()
        self.assertEqual(
            experiments,
            {
                models.Experiment(
                    name='switch_syntax',
                    identity='random',
                    variants=('on', 'off'),
                    seed='switch_syntax',
                    audiences=(
                        models.Audience(
                            rule=None,
                            allocations=(
                                models.Allocation(
                                    variant='on',
                                    percent=100,
                                ),
                            ),
                        ),
                    ),
                ),
                models.Experiment(
                    name='experimentitious',
                    identity='user.id',
                    variants=('active', 'inactive', 'control'),
                    seed='maseed',
                    audiences=(
                        models.Audience(
                            rule='locale == "en-US"',
                            allocations=(
                                models.Allocation(
                                    variant='active',
                                    percent=10,
                                ),
                                models.Allocation(
                                    variant='inactive',
                                    percent=80,
                                ),
                                models.Allocation(
                                    variant='control',
                                    percent=10,
                                ),
                            ),
                        ),
                        models.Audience(
                            rule=None,
                            allocations=(
                                models.Allocation(
                                    variant='inactive',
                                    percent=100,
                                ),
                            ),
                        ),
                    ),
                ),
            },
        )



from abc import ABC, abstractmethod
from operator import itemgetter
from typing import Iterable

from django.http import HttpRequest
from django.utils.module_loading import import_string

from cravensworth.core.conf import get_setting
from cravensworth.core.models import Allocation, Audience, Context, Experiment
from cravensworth.core.utils import get_tracking_key


class ContextProvider(ABC):
    """
    Base class for context providers.

    Context providers populate contexts.
    """

    @abstractmethod
    def context(self, **kwargs) -> Context:
        """
        Constructs a context.
        """
        raise NotImplementedError()


class DjangoRequestContextProvider(ContextProvider):
    """
    A context provider for Django environments.

    Pre-populates the context with the Django user (`user`) and tracking key for
    identifying anonymous users (`anonymous`).
    """

    def context(self, **kwargs) -> Context:
        """
        Constructs a `DjangoRequestContextProvider`.

        Args:
            **kwargs (dict[str, Any]): Arbitrary keyword arguments.
                - request (django.http.HttpRequest): Django request. Must be
                  passed as a keyword argument.
        """
        request = kwargs.get('request')
        if request is None or not isinstance(request, HttpRequest):
            raise ValueError(
                'Django request is required as a keyword argument to context()'
            )

        return Context(
            {
                'user': getattr(request, 'user', None),
                'anonymous': get_tracking_key(request),
            }
        )


DEFAULT_CONTEXT_PROVIDER = (
    'cravensworth.core.providers.DjangoRequestContextProvider'
)


def get_context_provider() -> ContextProvider:
    """
    Load and instantiate an instance of a context provider.
    """
    return import_string(
        get_setting('CONTEXT_PROVIDER', DEFAULT_CONTEXT_PROVIDER),
    )()


class Source(ABC):
    """
    Protocol for experiment sources.

    Classes implementing this protocol must provide a `load` method that returns
    an iterable of `Experiment` instances. This allows flexible sourcing of
    experiments from different backends (e.g., settings, database, API).
    """

    @abstractmethod
    def load(self) -> Iterable[Experiment]:
        """
        Loads all experiments that the project should be aware of.

        For projects that use cravensworth_middleware, `load()` will be called
        for every request.
        """
        raise NotImplementedError()


class SettingsSource(Source):
    """
    A source that loads experiments from a Django settings module.

    Expects the setting `EXPERIMENTS` to be a list of experiment specifications,
    each a dictionary containing:

    - `name` - The unique name of the experiment.
    - `identity` - The name of the context value to use as identity.
    - `variants` - A list of variant supported by the experiment.
    - `audiences` - A list of audiences and allocations for each.
    - `seed` (optional) - A seed for extra control over user hashing.

    Example experient spec:

        CRAVENSWORTH = {
            'EXPERIMENTS': [
                {
                    'name': 'super_cool_experiment',
                    'identity': 'random',
                    'variants': [
                        {'name': 'active'},
                        {'name': 'inactive'},
                        {'name': 'control'},
                    ],
                    'audiences': [
                        {
                            'rule': 'language == "en"',
                            'allocations': [
                                {'variant': 'active', 'percent': 50},
                                {'variant': 'inactive', 'percent': 0},
                                {'variant': 'control', 'percent': 50},
                            ],
                        },
                        {
                            'rule': None,
                            'allocations': [
                                {'variant': 'inactive', 'percent': 100},
                            ],
                        },
                    ],
                }
            ]
        }

    For instances where you want to use an experiment as a simple switch (always
    on or off for all users), this source provides some syntactic sugar to make
    defining switches easier:

        CRAVENSWORTH = {
            'EXPERIMENTS': [
                'cool_switch:on',
                'uncool_switch:off'
            ]
        }

    The above is equivalent to:

        CRAVENSWORTH = {
            'EXPERIMENTS': [
                {
                    'name': 'cool_switch',
                    'identity': 'random',
                    'variants': [
                        {'name': 'on'},
                        {'name': 'off'},
                    ],
                    'audiences': [
                        {
                            'rule': None,
                            'allocations': [
                                {'variant': 'on', 'percent': 100},
                            ],
                        },
                    ],
                }
            ]
        }
    """

    def load(self) -> Iterable[Experiment]:
        experiments = set()
        for experiment in get_setting('EXPERIMENTS', []):
            if type(experiment) is str:
                experiments.add(self._read_switch(experiment))
            else:
                experiments.add(self._read_object(experiment))
        return experiments

    @staticmethod
    def _read_object(experiment) -> Experiment:
        variants = tuple(
            itemgetter('name')(variant) for variant in experiment['variants']
        )
        audiences = []

        for audience in experiment['audiences']:
            rule = audience.get('rule')
            allocations = tuple(
                Allocation(
                    *itemgetter('variant', 'percent')(allocation),
                ).validate()
                for allocation in audience['allocations']
            )
            audiences.append(Audience(rule, allocations).validate())

        return Experiment(
            name=experiment['name'],
            identity=experiment['identity'],
            variants=variants,
            audiences=tuple(audiences),
            seed=experiment.get('seed'),
        ).validate()

    @staticmethod
    def _read_switch(value) -> Experiment:
        name, variant = value.rsplit(':', maxsplit=1)
        return Experiment(
            name,
            identity='random',
            variants=('on', 'off'),
            audiences=(
                Audience(
                    rule=None,
                    allocations=(Allocation(variant, percent=100).validate(),),
                ).validate(),
            ),
        ).validate()


def get_source() -> Source:
    return import_string(
        get_setting('SOURCE', 'cravensworth.core.providers.SettingsSource'),
    )()


__all__ = [
    'ContextProvider',
    'DjangoRequestContextProvider',
    'get_context_provider',
    'Source',
    'SettingsSource',
    'get_source',
]



from django.test import RequestFactory, SimpleTestCase, override_settings

from cravensworth.core.testing import override_experiment
from cravensworth.core.experiment import is_on


class TestOverrideExperiment(SimpleTestCase):
    factory = RequestFactory()

    @override_settings(CRAVENSWORTH={'EXPERIMENTS': ['switch:off']})
    def test_single_override(self):
        request = self.factory.get('/path')
        override_experiment(request, 'switch', 'on')

        self.assertTrue(is_on(request, 'switch'))

    @override_settings(
        CRAVENSWORTH={'EXPERIMENTS': ['switch1:off', 'switch2:on']}
    )
    def test_multiple_overrides(self):
        request = self.factory.get('/path')
        override_experiment(
            request,
            {
                'switch1': 'on',
                'switch2': 'off',
            },
        )

        self.assertTrue(is_on(request, 'switch1'))
        self.assertFalse(is_on(request, 'switch2'))

    def test_single_override_non_none_variant(self):
        request = self.factory.get('/path')

        with self.assertRaises(ValueError):
            override_experiment(request, 'switch', None)

    def test_multiple_overrides_non_dict_experiment(self):
        request = self.factory.get('/path')

        with self.assertRaises(TypeError):
            override_experiment(request, set('switch:on'))



import random
import re
from collections import UserDict
from dataclasses import dataclass
from typing import ClassVar, Mapping, Protocol, TypeVar

from rapidhash import rapidhash
from simpleeval import EvalWithCompoundTypes


class Context(UserDict):
    """
    Context contains contextual data for use by experiments in determining
    matching variants.
    """

    def __init__(self, data: dict = {}):
        super().__init__(data)
        self._identities = dict()

    def identity(self, keypath: str, seed: str | None) -> int:
        """
        Uses context data to calculate an identity for the given keypath and
        seed.

        Args:
            keypath (str): A path to specifying the key of the value in the
                context that is to be used as the identity. Paths are in object
                notation (period-delimited).
            seed (str, optional): A seed value that is combined with the key to
                produce an identity.

        `keypath` has a special value, "random", that will return a random value
        for use as the identity. Seed has no effect if "random" is used as the
        keypath.

        Identity values are cached so, the same keypath/seed pair will not be
        re-calculated if identity() is called again.

        Raises:
            KeyError: If the keypath does not exist in the context or its
                corresponding value is None.
        """
        cachekey = f'{keypath}{seed or ""}'
        if cachekey not in self._identities:
            self._identities[cachekey] = self._calculate_identity(keypath, seed)
        return self._identities[cachekey]

    def _calculate_identity(self, keypath, seed) -> int:
        if keypath == 'random':
            return random.randint(0, 99)
        identity = self._get_key_by_path(self, keypath)
        if identity is None:
            raise KeyError(
                f'Identity keypath "{keypath}" not found in the context, or the'
                'value is None'
            )
        return rapidhash(f'{identity}{seed}'.encode()) % 100

    @staticmethod
    def _get_key_by_path(obj, path):
        keys = path.split('.')
        current = obj
        for key in keys:
            if isinstance(current, Mapping):
                current = current.get(key)
            else:
                current = getattr(current, key, None)
            if current is None:
                break
        return current


_symbol_pattern = re.compile(r'^[\w\.]+$', re.ASCII)
_name_pattern = re.compile(r'^[\w]+$', re.ASCII)


T = TypeVar('T', bound='_Validatable')


class _Validatable(Protocol):
    def validate(self: T) -> T:
        for name, _ in self.__dataclass_fields__.items():
            validate = getattr(self, f'validate_{name}', None)
            if validate is not None and callable(validate):
                validate()
        return self


@dataclass(frozen=True, eq=True)
class Allocation(_Validatable):
    """
    Allocation represents the portion of an audience that is allocated to a
    particular variant.
    """

    variant: str
    percent: int

    def validate_variant(self):
        if not _name_pattern.match(self.variant):
            raise ValueError('Variant must contain only [a-zA-Z0-9_]')

    def validate_percent(self):
        if self.percent < 0:
            raise ValueError('Percent must not be negative')
        if self.percent > 100:
            raise ValueError('Percent must not be greater than 100')


@dataclass(frozen=True, eq=True)
class Audience(_Validatable):
    """
    An audience is a population of entities that all share a matching set of
    characteristics or, in the case of a default rule, no matching
    characteristics.

    Entities within an audience are assigned to an allocation based on their
    respective identities.
    """

    rule: str | None
    allocations: tuple[Allocation]
    evaluator: ClassVar[EvalWithCompoundTypes] = EvalWithCompoundTypes()

    def __post_init__(self):
        if self.rule is not None:
            try:
                object.__setattr__(
                    self, '_rule_parsed', self.evaluator.parse(self.rule)
                )
            except SyntaxError as e:
                raise ValueError('Invalid rule syntax') from e

    def validate_allocations(self):
        percent_total = 0
        for allocation in self.allocations:
            allocation.validate()
            percent_total += allocation.percent

        if percent_total != 100:
            raise ValueError('Allocations must sum to 100 percent')

    def matches(self, context: Context) -> bool:
        """
        Tests an entity to see if it matches the rule for inclusion in the
        audience.
        """
        if self.rule is None:
            return True
        self.evaluator.names = context
        result = self.evaluator.eval(self.rule, self._rule_parsed)
        if type(result) is not bool:
            raise TypeError('Audience rule must evaluate to a boolean value')
        return result

    def determine_variant(self, rangekey: int) -> str:
        """
        Determines the variant that matches a given entity based on the position
        of its identity within the range of allocations within this audience.
        """
        range_start = -1
        for allocation in self.allocations:
            range_end = range_start + allocation.percent
            if range_start < rangekey <= range_end:
                return allocation.variant
            range_start = range_end


@dataclass(frozen=True, eq=True)
class Experiment(_Validatable):
    """
    Experiment represents a test that can be used to verify a hypothesis by
    bucketing entities into multiple variants.
    """

    name: str
    identity: str
    variants: tuple[str]
    audiences: tuple[Audience]
    seed: str | None = None

    def __post_init__(self):
        if self.seed is None:
            object.__setattr__(self, 'seed', self.name)

    def validate_name(self):
        if not _name_pattern.match(self.name):
            raise ValueError('Name must contain only [a-zA-Z0-9_]')

    def validate_variants(self):
        if len(self.variants) < 1:
            raise ValueError('Experiment must define at least one variant')

    def validate_identity(self):
        if self.identity == 'random':
            return True
        if not _symbol_pattern.match(self.identity):
            raise ValueError('Invalid identity symbol name "{self.identity}"')

    def validate_audiences(self):
        if len(self.audiences) == 0:
            raise ValueError('Experiment must define at least one audience')
        if any(audience.rule is None for audience in self.audiences[:-1]):
            raise ValueError('Only the last audience rule can be None')
        if self.audiences[-1].rule is not None:
            raise ValueError('Last audience must not define a rule')

        for audience in self.audiences:
            for allocation in audience.allocations:
                allocation.validate()
                if allocation.variant not in self.variants:
                    raise ValueError(
                        f'Undeclared variant "{allocation.variant}"',
                    )

    def determine_variant(self, context: Context, override: str | None) -> str:
        """
        Determines which variant an entity should use by matching them against
        audience rules.

        Audiences will be matched in the order in which they are defined.
        """
        if override is not None and override in self.variants:
            return override

        for audience in self.audiences:
            if audience.matches(context):
                identity = context.identity(self.identity, self.seed)
                return audience.determine_variant(identity)


__all__ = [
    'Context',
    'Allocation',
    'Audience',
    'Experiment',
]



from typing import Any, ParamSpec, TypeVar, Callable

from django.http import Http404
from django.shortcuts import redirect

from cravensworth.core import experiment


P = ParamSpec('P')
R = TypeVar('R')


def variant(
    name: str, variant: str | list[str], redirect_to: Any = None
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Controls access to the decorated view based on the active variant of the
    given experiment.

    If the given variant or variants are active, the view will be executed. If
    the variant is not a match, the user agent will be redirected to the
    specified redirect target, or will receive a 404 if no redirect is provided.

    Args:
        name (str): The name of the experiment.
        variant (str | list[str]): The name of the variant(s) that control
            access the view. Can be a single string (e.g., 'on') or a list of
            strings (e.g., `['variant_a', 'variant_b']`).
        redirect_to (Any, optional): If provided, the request will be redirected
            to this target if the active variant does not match the specified
            variant(s). If `None`, an `Http404` will be raised.

    `redirect_to` can be any type allowed by Django redirects:

    - A view name. `reverse()` will be used to get the redirect URL. Arguments
      are not supported.
    - A model that defines `get_absolute_url`.
    - An absolute or relative URL.

    Redirects are temporary (302).

    Returns:
        function: The decorated view function.

    Raises:
        Http404: If the active variant does not match the specified variant(s)
            and `redirect_to` is not provided.
    """

    def decorator(function: Callable[P, R]) -> Callable[P, R]:
        def wrapper(request, *args: P.args, **kwargs: P.kwargs) -> R:
            if experiment.is_variant(request, name, variant):
                return function(request, *args, **kwargs)
            elif redirect_to is not None:
                return redirect(redirect_to)
            raise Http404

        return wrapper

    return decorator


def switch_on(
    name: str, redirect_to: Any = None
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Controls access to the decorated view based on the active variant of the
    given experiment.

    If the active variant is 'on', the view will be executed. Otherwise the user
    agent will be redirected to the specified redirect target, or will receive
    a 404 if no redirect is provided.

    Args:
        name (str): The name of the experiment.
        redirect_to (Any, optional): If provided, the request will be redirected
            to this target if the active variant does not match the specified
            variant(s). If `None`, an `Http404` will be raised.

    `redirect_to` can be any type allowed by Django redirects:

    - A view name. `reverse()` will be used to get the redirect URL. Arguments
      are not supported.
    - A model that defines `get_absolute_url`.
    - An absolute or relative URL.

    Redirects are temporary (302).

    Returns:
        function: The decorated view function.

    Raises:
        Http404: If the experiment is off and `redirect_to` is not provided.
    """
    return variant(name, 'on', redirect_to)


def switch_off(
    name: str, redirect_to: Any = None
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Controls access to the decorated view based on the active variant of the
    given experiment.

    If the active variant is 'off', the view will be executed. Otherwise the
    user agent will be redirected to the specified redirect target, or will
    receive a 404 if no redirect is provided.

    Args:
        name (str): The name of the experiment.
        redirect_to (Any, optional): If provided, the request will be redirected
            to this target if the active variant does not match the specified
            variant(s). If None (default), an Http404 will be raised.

    `redirect_to` can be any type allowed by Django redirects:

    - A view name. `reverse()` will be used to get the redirect URL. Arguments
      are not supported.
    - A model that defines `get_absolute_url`.
    - An absolute or relative URL.

    Redirects will be temporary only (302).

    Returns:
        function: The decorated view function.

    Raises:
        Http404: If the experiment is on and `redirect_to` is not provided.
    """
    return variant(name, 'off', redirect_to)



from django.apps import AppConfig


class CravensworthCoreConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'cravensworth.core'
    verbose_name = 'Django Cravensworth'
    label = 'cravensworth_core'
